#include<iostream>
#include<fstream>
#include<cstdlib>
#include<cstring>
#include<getopt.h>
#include<stdint.h>
#include<libgen.h>
#include<seqan/sequence.h>
#include<seqan/file.h>

using namespace seqan;

#define SEQMIX_none   0
#define SEQMIX_Mto1   1
#define SEQMIX_MtoN   2
#define SEQMIX_random 3

static bool verbose = 0;
const char *optstring = "hvm:o:l:b:f:d:rp:n:L:c";

int seqmix_MtoN(void);
int seqmix_Mto1(struct Mto1_params *params);
int seqmix_random(struct random_params *params);
int seqgen_rnd(size_t n, void *pp);

void print_usage(const char *cmd)
{
        ::std::cout << ::std::endl;
        ::std::cout << "Usage: " << cmd << " [-vh] -l <logfile> -o <outfile> <infile> ..." << ::std::endl;
        ::std::cout << ::std::endl;
        ::std::cout << '\t' << "-h                  - this help"  << ::std::endl;
        ::std::cout << '\t' << "-v                  - be verbose" << ::std::endl;
        ::std::cout << '\t' << "-m <n>              - run mode: 1 => M-to-1" << ::std::endl;
        ::std::cout << '\t' << "                                2 => M-to-N" << ::std::endl;
        ::std::cout << '\t' << "                                3 => random" << ::std::endl;
        ::std::cout << ::std::endl;

        /*=== M-to-1 mode options ===*/
        ::std::cout << '\t' << "-------- M-to-1 mode options -------" << ::std::endl << ::std::endl;
        ::std::cout << '\t' << "-o <file>           - write output to file <file>" << ::std::endl;
        ::std::cout << '\t' << "-l <file>           - write log to file <file>" << ::std::endl;
        ::std::cout << '\t' << "-b <file>           - background sequence(s) (if not specified, random sequence" << ::std::endl;
        ::std::cout << '\t' << "                      will be generated)" << ::std::endl;
        ::std::cout << '\t' << "-f <n>              - frequency of sequences from the input file in the randomly" << ::std::endl;
        ::std::cout << '\t' << "                      generated sequence or in the background sequence"  << ::std::endl;
        ::std::cout << '\t' << "-d <n>              - maximal deviation from the specified frequency"    << ::std::endl;
        ::std::cout << '\t' << "-r                  - randomize the order of input sequences in the output sequence(s)" << ::std::endl;
        ::std::cout << '\t' << "-p <a>,<t>,<c>,<g>  - comma separated list of probabilities of the nucleotides in"  << ::std::endl;
        ::std::cout << '\t' << "                      the random sequence. The sum of these numbers must be equal" << ::std::endl;
        ::std::cout << '\t' << "                      to one. The numbers are expected to be in the range <0.0;1.0>." << ::std::endl;
        ::std::cout << '\t' << "-c                  - cycle through the input sequences until the maximum length is" << ::std::endl;
        ::std::cout << '\t' << "                      reached. This option requires the -L (maximal length) option" << ::std::endl;
        ::std::cout << '\t' << "                      to be specified, see bellow." << ::std::endl;
        ::std::cout << '\t' << "-L <n>              - maximal length of the output sequence" << ::std::endl;
        ::std::cout << ::std::endl;

        /*=== M-to-N mode options ===*/
        ::std::cout << '\t' << "-------- M-to-N mode options -------" << ::std::endl << ::std::endl;
        ::std::cout << '\t' << " Sorry, not implemented yet." << ::std::endl;
        ::std::cout << ::std::endl;

        /*=== random mode options ===*/
        ::std::cout << '\t' << "-------- random mode options -------" << ::std::endl << ::std::endl;
        ::std::cout << '\t' << "-o <file>           - write output to file <file>" << ::std::endl;
        ::std::cout << '\t' << "-n <n>              - generate <n> random sequences (default: 1)" << ::std::endl;
        ::std::cout << '\t' << "-L <n>              - length of the randomly generated sequence(s)" << ::std::endl;
        ::std::cout << '\t' << "-d <n>              - maximal deviation from the specified length"    << ::std::endl;
        ::std::cout << '\t' << "-p <a>,<t>,<c>,<g>  - comma separated list of probabilities of the nucleotides in"  << ::std::endl;
        ::std::cout << '\t' << "                      the random sequence. The sum of these numbers must be equal" << ::std::endl;
        ::std::cout << '\t' << "                      to one. The numbers are expected to be in the range <0.0;1.0>." << ::std::endl;

        ::std::cout << ::std::endl;
}

struct seqFile {
        char   *filename; /* filename */
        FILE   *sqstream; /* file stream */
        size_t  raw_size; /* raw size in bytes */
        size_t  num_seqs; /* sum of all lengths of sequences */
        size_t  max_seql; /* longest sequence lenght */
};

typedef int (seqgen_t)(size_t, void *);

struct Mto1_params {
        struct seqFile *infiles;
        uint32_t        infiles_cnt;

        FILE *outfile;
        FILE *logfile;

        seqgen_t *seqgen;
        void     *seqgen_ptr;

        size_t   frequency;
        size_t   deviation;
        size_t   max_length;

        int      cycle_seqs;
};

struct random_params {
        FILE     *outfile;

        int       num_seqs;
        size_t    max_length;
        size_t    deviation;

        seqgen_t *seqgen;
        void     *seqgen_ptr;
};

#define A 0
#define T 1
#define C 2
#define G 3

#define FASTA_LINE_WIDTH 80

int main(int argc, char *argv[])
{
        char *arg0 = argv[0];
        int   opt;

        struct seqFile *infiles = NULL;
        uint32_t        infiles_cnt = 0;

        int    runmode = 0;
        char   *outfile = NULL;
        FILE   *outstream;
        char   *logfile = NULL;
        FILE   *logstream;
        size_t frequency  = 1000;
        size_t deviation  = 0;
        size_t max_length = 0;

        double p[4] = { /* A */ 0.25,
                        /* T */ 0.25,
                        /* C */ 0.25,
                        /* G */ 0.25 };

        int randomize  = 0;
        int num_seqs   = 1;
        int cycle_seqs = 0;

        srandom(getpid());

        while ((opt = getopt(argc, argv, optstring)) != -1) {
                switch (opt) {
                case 'h':
                        print_usage(basename(arg0));
                        return(-1);
                case 'v':
                        verbose = 1;
                        break;
                case 'm':
                        runmode = atoi(optarg);

                        switch(runmode) {
                        case SEQMIX_Mto1:
                        case SEQMIX_MtoN:
                        case SEQMIX_random:
                                break;
                        default:
                                ::std::cout << "Invalid run mode!" << ::std::endl;
                                return(-1);
                        }
                        break;
                case 'o':
                        outfile = strdup(optarg);
                        break;
                case 'l':
                        logfile = strdup(optarg);
                        break;
                case 'b':
                        //bgseq.filename = strdup(optarg);
                        break;
                case 'f':
                        frequency = atoll(optarg);
                        break;
                case 'd':
                        deviation = atoll(optarg);
                        break;
                case 'r':
                        randomize = 1;
                        break;
                case 'p': {
                        double p_S = 0;
                        char *p_tok;

                        p[A] = p[T] = p[G] = p[C] = 0.0;

                        do {
                                if ((p_tok = strtok(optarg, ",")) == NULL)
                                        break;
                                p[A] = atof(p_tok);

                                if ((p_tok = strtok(NULL, ",")) == NULL)
                                        break;
                                p[T] = atof(p_tok);

                                if ((p_tok = strtok(NULL, ",")) == NULL)
                                        break;
                                p[C] = atof(p_tok);

                                if ((p_tok = strtok(NULL, ",")) == NULL)
                                        break;
                                p[G] = atof(p_tok);
                        } while (0);

                        p_S  = p[A] + p[T] + p[C] + p[G];
                        p[A] = p[A] == 0 ? 1 - p_S: p[A];
                        p[T] = p[T] == 0 ? 1 - p_S: p[T];
                        p[C] = p[C] == 0 ? 1 - p_S: p[C];
                        p[G] = p[G] == 0 ? 1 - p_S: p[G];
                        p_S  = p[A] + p[T] + p[C] + p[G];

                        if (p_S < 0.9999 || p_S > 1.0001) {
                                ::std::cout << "Invalid probability string, sum of probabilities != 1.0" << ::std::endl;
                                return(-1);
                        }

                        break;
                }
                case 'n':
                        num_seqs = atoi(optarg);

                        if (num_seqs < 1) {
                                ::std::cout << "Invalid number of output sequences specified. The minimum is 1 sequence." << ::std::endl;
                                return(-1);
                        }

                        break;
                case 'L':
                        max_length = atoll(optarg);
                        break;
                case 'c':
                        cycle_seqs = 1;
                        break;
                default:
                        print_usage(basename(arg0));
                        return(-1);
                }
        }

        argc -= optind;
        argv += optind;

        if (outfile == NULL) {
                ::std::cout << "Please specify the output file using the `-o' option" << ::std::endl;
                return(-1);
        }

        outstream = fopen(outfile, "wb");

        if (outstream == NULL) {
                ::std::cout << "Can't open output file: " << outfile << ::std::endl;
                return(-1);
        }

        switch (runmode) {
        case SEQMIX_Mto1:
                if (cycle_seqs) {
                        if (max_length == 0) {
                                ::std::cout << "Option `-c' requires option `-L' to be specified." << ::std::endl;
                                return(-1);
                        }
                } else {
                        max_length = ~((size_t)0);
                }
        case SEQMIX_MtoN: {
                if (deviation > frequency) {
                        ::std::cout << "Deviation cannot be larger than frequency!" << ::std::endl;
                        return(-1);
                }

                infiles_cnt = argc;
                infiles     = (struct seqFile *)malloc(sizeof(*infiles) * infiles_cnt);

                seqan::String<seqan::Iupac> tmpstr;
                size_t                      tmplen;

                for (size_t i = 0; i < infiles_cnt; ++i) {
                        infiles[i].filename = argv[i];

                        if (verbose)
                                ::std::cout << "[" << i+1 << "/" << infiles_cnt << "] Opening " << argv[i] << ::std::endl;

                        infiles[i].sqstream = fopen(argv[i], "rb");

                        if (infiles[i].sqstream == NULL) {
                                ::std::cout << "Can't open input file: " << infiles[i].filename << ::std::endl;
                                return(-1);
                        }

                        infiles[i].num_seqs = 0;
                        infiles[i].max_seql = 0;
                        infiles[i].raw_size = 0;

                        while(!feof(infiles[i].sqstream)) {
                                read(infiles[i].sqstream, tmpstr, seqan::Fasta());
                                tmplen = length(tmpstr);
                                ++(infiles[i].num_seqs);

                                if (tmplen > infiles[i].max_seql)
                                        infiles[i].max_seql = tmplen;
                        }

                        rewind(infiles[i].sqstream);
                        clearerr(infiles[i].sqstream);
                }
        }
                break;
        case SEQMIX_random:
                if (deviation > max_length) {
                        ::std::cout << "Deviation cannot be larger than the maximal length!" << ::std::endl;
                        return(-1);
                }
                break;
        }

        switch (runmode) {
        case SEQMIX_Mto1: {
                struct Mto1_params params;

                if (logfile == NULL) {
                        ::std::cout << "Please specify the logfile using the `-l' option" << ::std::endl;
                        return(-1);
                }

                logstream = fopen(logfile, "w");

                if (logstream == NULL) {
                        ::std::cout << "Can't open log file: " << logfile << ::std::endl;
                        return(-1);
                }

                params.infiles     = infiles;
                params.infiles_cnt = infiles_cnt;
                params.outfile     = outstream;
                params.logfile     = logstream;

                params.seqgen      = seqgen_rnd;
                params.seqgen_ptr  = &p;

                params.frequency   = frequency;
                params.deviation   = deviation;
                params.max_length  = max_length;
                params.cycle_seqs  = cycle_seqs;

                return seqmix_Mto1(&params);
        }
        case SEQMIX_MtoN: {
                return seqmix_MtoN();
        }
        case SEQMIX_random: {
                struct random_params params;

                params.outfile    = outstream;
                params.num_seqs   = num_seqs;
                params.max_length = max_length;
                params.deviation  = deviation;
                params.seqgen     = seqgen_rnd;
                params.seqgen_ptr = &p;

                return seqmix_random(&params);
        }
        default:
                ::std::cout << "Run mode not specified! Use the `-m' option to do so." << ::std::endl;
        }

        return(-1);
}

int seqmix_MtoN(void)
{
        return(-1);
}

int seqmix_Mto1(struct Mto1_params *params)
{
        struct seqFile *seqfile;
        int32_t bgn;
        size_t  pos;

        seqan::String<Iupac> seqstr;
        seqan::String<char>  seqhdr;

        if (verbose) {
                ::std::cout << "Generating sequence" << ::std::endl;
        }

        fprintf(params->outfile, ">Sequence_1\n");
        fprintf(params->logfile, "# file#:seq#\tpos\tlen\n");
        pos = 0;

        do {
                for (size_t fn = 0; fn < params->infiles_cnt && params->max_length > 0; ++fn) {
                        seqfile = params->infiles + fn;

                        if (verbose) {
                                ::std::cout << "Reading from " << seqfile->filename << ::std::endl;
                        }

                        for (size_t sn = 0; sn < seqfile->num_seqs; ++sn) {
                                bgn = (int32_t)params->frequency + (params->deviation > 0 ?
                                                                    (int32_t)((random() % (params->deviation << 1)) - params->deviation) : 0);

                                if (bgn < 0)
                                        bgn = 0;

                                if ((size_t)bgn > params->max_length)
                                        bgn = params->max_length;

                                if (verbose) {
                                        ::std::cout << "bgn= " << bgn << ::std::endl;
                                }

                                pos += bgn;
                                params->max_length -= bgn;

                                while (bgn > 0) {
                                        fputc(params->seqgen(0, params->seqgen_ptr), params->outfile);
                                        --bgn;
                                }

                                readMeta(seqfile->sqstream, seqhdr, seqan::Fasta());
                                read(seqfile->sqstream, seqstr, seqan::Fasta());
                                write(params->outfile, seqstr);

                                if (length(seqstr) < params->max_length)
                                        params->max_length -= length(seqstr);
                                else
                                        params->max_length  = 0;

                                fprintf(params->logfile, "%zu:%zu\t%zu\t%zu # F: %s, H: %s\n",
                                        fn, sn, pos, length(seqstr), seqfile->filename, toCString(seqhdr));

                                pos += length(seqstr);
                        }

                        rewind(seqfile->sqstream);
                        clearerr(seqfile->sqstream);
                }
        } while (params->max_length > 0 && params->cycle_seqs);

        fclose(params->outfile);
        fclose(params->logfile);

        return(0);
}

int seqmix_random(struct random_params *params)
{
        int32_t len;

        for (int n = 0; n < params->num_seqs; ++n) {
                fprintf(params->outfile,
                        ">Random sequence %d; p(A,T,C,G)=(%.2f,%.2f,%.2f,%.2f)\n", n,
                        ((double *)(params->seqgen_ptr))[A],
                        ((double *)(params->seqgen_ptr))[T],
                        ((double *)(params->seqgen_ptr))[C],
                        ((double *)(params->seqgen_ptr))[G]);

                len = params->max_length + (params->deviation > 0 ?
                                            (int32_t)((random() % (params->deviation << 1)) - params->deviation) : 0);

                while (len > 0) {
                        fputc(params->seqgen(0, params->seqgen_ptr), params->outfile);
                        --len;
                }

                fputc('\n', params->outfile);
        }

        fclose(params->outfile);

        return(0);
}

const char *_atcg = "ATCG";

int seqgen_rnd(size_t n, void *pp)
{
        register int i;
        double *p = (double *)(pp);
        double rnd;

        rnd = ((double)(random()))/((double)(RAND_MAX));

        for (i = 0; i < 4 && rnd > p[i]; ++i)
                rnd -= p[i];

        i = i > 3 ? 3 : i;

        return (int)(*(_atcg + i));
}
